!=========================================================================
! Copyright (C) GemTalk Systems 1986-2013.  All Rights Reserved.
!
! $Id: bomlastconv.gs,v 1.12 2008-03-04 22:48:29 bretlb Exp $
!
! Description - 
!   This file contains the subset of bomlast.gs used for repository
!   conversion. (bomlast.gs is not used for conversion).
!=========================================================================

! Fix up methods that would have contained forward reference to PrintStream.
!  PrintStream requires execution to create class; execution not
!  possible until filein2 stage, but we need Object>>printString in 
!  filein1 in order to make topaz execution work ... 
category: 'Formatting'
method: Object
printString

"Returns a String whose contents are a displayable representation of the
 receiver."

"This method uses the printOn: method to create the String."

| ws str |

str := String new.

ws := PrintStream printingOn: str.  

self printOn: ws.
^ ws _collection "contents might have been converted to an EUCString."
%


! compile methods that contain forward references to Float if they are compiled
! earlier in the filein when building a fresh GemStone repository.

category: 'Converting'
method: Integer
asDecimalFloat

"Returns a DecimalFloat representing the receiver."

<primitive: 137>

| factor sum oldRoundingMode |

oldRoundingMode := Float roundingMode.  "nearest even is only right way"
Float roundingMode: 'nearestEven'.

sum := 0.0F0 .
factor := self sign asFloat.
1 to: self size do:
  [:i |
  sum := (self _digitAt: i) * factor + sum.
  factor := factor * 32768.0F0 ].

DecimalFloat roundingMode: oldRoundingMode.
^sum
%

category: 'Truncation and Rounding'
method: Number
ceiling

"Returns the integer that is closest to the receiver, on the same side
 of the receiver as positive infinity.  In particular, returns the
 receiver if the receiver is an integer."

(self <= 0.0)
   ifTrue: [^self truncated]
   ifFalse: [^self negated floor negated truncated]
%

category: 'Arithmetic'
method: Number
degreesToRadians

"Assuming the receiver represents an angle in degrees, returns the angle
 in radians."

^ self / 57.29577951308232
%

category: 'Arithmetic'
method: Number
radiansToDegrees

"Assuming the receiver represents an angle in radians, returns the angle
 in degrees."

^ self * 57.29577951308232
%

method: GsMethodDictionary
rebuildTable: newSize

"Rebuilds the method dictionary by populating a larger method dictionary
 first and doing a (primitive) become:"

"This is a reimplementation of the method in object.gs to handle
instances that have a dependency list."

<primitive: 901>
| newGsMethodDict depList |

tableSize = newSize ifTrue:[ 
  System _disableProtectedMode.
  ^ self "no change in table size"
].
collisionLimit == 536870911 ifTrue:[
  "rehash disabled during symbol canonicalization."
  System _disableProtectedMode.
  ^ self
  ].

newGsMethodDict := self class new: (newSize * 2).
newGsMethodDict valueConstraint: valueConstraint.
newGsMethodDict keyConstraint: keyConstraint.

self keysAndValuesDo: [ :aKey :aValue |
  newGsMethodDict at: aKey put: aValue.
  ].

(depList := DependencyList for: self) ~~ nil
  ifTrue: [ DependencyList set: nil for: self ].

newGsMethodDict _primitiveBecome: self.

depList ~~ nil
  ifTrue: [ DependencyList set: depList for: self ].
System _disableProtectedMode.
%

!--------------------------------------
! build global float literals for exceptional floats

run
#[ #PlusInfinity , #MinusInfinity, #PlusQuietNaN,
   #MinusQuietNaN, #PlusSignalingNaN, #MinusSignalingNaN ]
 do: [ :aSymbol |
    ( Globals at: aSymbol ifAbsent: [ nil ] ) == nil
      ifTrue: [
        Globals at:aSymbol put: (Float fromString: aSymbol) .
        Globals at: (Symbol withAll: 'Decimal' + aSymbol)
                put: (DecimalFloat fromString: aSymbol) .
      ]
    ] .
^ true
%

!--------------------------------------

category: 'Repository Conversion'
method: Class
_unsafeDisallowSubclasses

""

self _unsafeAt: 2 "offset of format"
           put: (format bitOr: 32 "subclasses disallowed bit") .
self _refreshClassCache
%

!--------------------------------------
run
| list |
  list := #[ Behavior, Class, Metaclass, Repository, UserProfileSet,
     AbstractUserProfileSet,
     CompiledMethod, Float, DecimalFloat, SmallFloat, Symbol,
     DoubleByteSymbol, 
     UnimplementedFloat1, UnimplementedFloat2 ,
     MethodContext, Block, SelectionBlock,
     Segment, SymbolList, System, UndefinedObject, UserProfile, Boolean,
     Character, SmallInteger, Integer, LargePositiveInteger,
     LargeNegativeInteger, JISCharacter, StackSegment,
     StackBuffer, Activation, Process, VariableContext, BlockClosure,
     ExecutableBlock, SimpleBlock, ComplexBlock, ComplexVCBlock,

     GsMethod, GsProcess, GsMethodDictionary, GsStackBuffer, 

     SelectBlock , BtreeBasicInteriorNode, BtreeBasicLeafNode, BasicSortNode,
     RcIndexBucket, RcIndexBucketWithCache, RcIndexDictionary, NscBuilder,

     PathTerm, ConstrainedPathTerm, SetValuedPathTerm  "fix bug 11484"
     ].

  list do: [ :aClass | aClass _unsafeDisallowSubclasses ] .

  list do: [ :aClass |
     aClass subclassesDisallowed ifFalse:[ ^ 'Error on ' + aClass name ].
     ] .
  ^ true
%

run
Class removeSelector: #_unsafeDisallowSubclasses .
^ true
%


run
| list blk |

" allow adding instance variables to the following classes.  fixes bug 6569"
list := #[ UserProfile ] .

blk := [ :aClass |
          aClass _unsafeAt: 4 "instVarNames"
                 put: (Array withAll: (aClass instVarAt: 4)) .
          aClass _unsafeAt: 5 "constraints"
                     put: (Array withAll: (aClass instVarAt: 5)) .
          aClass _refreshClassCache
       ].
list do: blk .
^ true
%

run

|newConstraints|

"Change the constraint on the unwind block of VariableContext.  Fixes bug15424."

newConstraints := Array with: ExecutableBlock.
newConstraints immediateInvariant.
VariableContext _unsafeAt: 5 "constraints"
   put: newConstraints.
VariableContext _refreshClassCache.
^true
%


!--------------------------------------

!--------------------------------------
! clear "conversion in progress flag"   ! was object 906 in 4.1 
expectvalue false
send @3625 at: 1 put: false 


!--------------------------------------
! Create class documentation for kernel classes that are only placeholders.

category: 'For Documentation Installation only'
classmethod: System
installPlaceholderDocumentation

| str |

str :=
' is a placeholder for another related GemStone product that has not currently
  been installed.' .

" list of placeholder classes for other products "
#[ GsfClassDefinitionInfo, GsfModificationLog ]
  do: [ :aClass |
    | doc txt |
    doc := GsClassDocumentation _newForPrivateGsClass: aClass.
    txt := (GsDocText new) details:
           ((String withAll: (aClass name)) add: str; yourself).
    doc documentClassWith: txt.
    aClass description: doc.
  ].

" list of placeholder classes within the server "
#[ UnimplementedFloat1, UnimplementedFloat2 ]
  do: [ :aClass |
    | doc |
    doc := GsClassDocumentation _newForPrivateGsClass: aClass.
    doc documentClassWith: (GsDocText new).
    aClass description: doc.
  ].
%
! helps debugging native code
!commit
!pausefordebug


!--------------------------------------
! Install class documentation for kernel classes only.

run

| classes baseNonKernelClassNames |

System installPlaceholderDocumentation.

baseNonKernelClassNames := #[#DateAndTimeANSI, #DateAndTime, #Duration, 
  #TimeZoneInfo, #TimeZoneTransition, #TimeZone, #SimpleTestResource,
  #ExampleSetTest, #SimpleTestResourceTestCase, #ResumableTestFailure, 
  #SUnitTest, #TestFailure, #SUnitDelay, #ResumableTestFailureTestCase, 
  #SUnitNameResolver, #TestResource, #TestCase, #TestResult, #TestSuite,
  #ExceptionSet, #ExceptionHandler, #ExceptionA, #FailedMessage, #Error, 
  #Notification, #Warning, #MessageNotUnderstood, #ZeroDivide].

classes := ClassSet new.
(Globals values) do: [ :each |
  ( (each isKindOf: Class)
      _and:[(each class includesSelector: #installDocumentation)
      _and: [each _isKernel 
             _or: [baseNonKernelClassNames includes: each thisClass name]]] )
  ifTrue: [ classes add: each.  ].
  ].

classes do: [ :aClass | aClass installDocumentation ] .
classes do: [ :aClass |
  (aClass class) removeCategory: 'For Documentation Installation only'.
  ].
^true
%

!------------------------------------------------------------
! Make various methods in the indexing subsystem 'protected'
!  now done via compilation directives during method compilation
!------------------------------------------------------------

!------------------------------------------------------------
! Make various classes in the indexing subsystem 'protected'.
! This prevents accessing an instance through structural access
!------------------------------------------------------------
!
run
" list of classes which will be protected "
#[ BasicSortNode, BtreeBasicInteriorNode, BtreeBasicLeafNode,
   BtreeInteriorNode, BtreeLeafNode, BtreeNode, BtreeReadStream,
   BucketValueBag, ConstrainedPathTerm, DependencyList, IdentityIndex,
   IndexList, PathEvaluator, PathSorter, PathTerm, QueryExecuter,
   RangeEqualityIndex, RcIndexBucket, RcIndexBucketWithCache, SegmentSet,
   SetValuedPathTerm, SortNode, SortNodeArray, RcIndexDictionary,
   ConstrainedPathEvaluator, SetValuedPathEvaluator, UserSecurityData,
]
    do: [ :aClass | aClass _makeProtected ].

^ true
%

!------------------------------------------------------------
! Build the dictionary of configuration file parameters
!------------------------------------------------------------
run
Globals at:#ConfigurationParameterDict put:
(System _configFileParameterDict) .
^ true
%
!------------------------------------------------------------
! Build the dictionary of software version information parameters
!------------------------------------------------------------
run
"dictionary to translate names to internal identifiers"
Globals at:#VersionParameterDict put:
(System _versionParameterDict) .
^ true
%
run
"dictionary of image version information"
Globals at:#ImageVersion put: SymbolDictionary new .
^ true
%
input $upgradeDir/version.topaz

! Fix 37168
! This segment used to be used for Timezones in 6.1.
! We need to create it here even though we won't use it
! so GS64 conversion works correctly.
run
| seg |
" create a new segment that is world writable "
seg := Segment newInRepository: SystemRepository.
seg worldAuthorization: #write.

" don't allow anyone else to change the authorization of the new segment "
seg changeToSegment: SystemSegment.
seg immediateInvariant.
Globals at: #GsTimeZoneSegment put: seg.
^ true
%
commit



! build the set of classes for which instances are disallowed
run
Globals at:#InstancesDisallowed put:
 ( IdentitySet withAll:
  #[ Activation, Character, Behavior, Block, BlockClosure, Boolean,
      Class, GsMethod, ComplexBlock, ComplexVCBlock, DateTime,
      ExecutableBlock, Float, DecimalFloat, SmallFloat, BinaryFloat,
      Integer, JISCharacter, LargeNegativeInteger,
      LargePositiveInteger, Metaclass, MethodContext, Number, Process,
      Repository, Segment, SelectBlock, SelectionBlock, SimpleBlock,
      SmallInteger, StackBuffer, StackSegment, System, UndefinedObject,
      UserProfile, VariableContext ]) .
^ true
%
commit

!------------------------------------------------------------
! Create collating tables for raw ASCII collating order for use with
! String compare methods  in category 'Other comparisons'
!------------------------------------------------------------
run
| a |
a := ByteArray new .
0 to: 255 do:[:j| a add: j ].
Globals at:#AsciiCollatingTable put: a .
^ a size = 256
%
run
| a |
a := DoubleByteString new .
0 to: 255 do:[:j| a add: (Character withValue: j )].
Globals at:#DoubleByteAsciiCollatingTable put: a .
^ a size = 256
%

!-----------
!  ensure compiler creates the canonicalized float zeros
run
| x |
x := 0.0 .
x class == Float ifFalse:[ ^ 'bad class for 0.0' ].
x := 0.0 asSmallFloat.
x class == SmallFloat ifFalse:[ ^ 'bad class for 0.0 asSmallFloat' ].
^ true
%
expectvalue @3745
run
0.0
%
expectvalue @3749
run
0.0 asSmallFloat
%

!-------------
! Add new Global as part of fix for bug 19391
expectvalue false
run
(Globals at: #AddSystemObjectsToGcCandidates otherwise: nil) == nil
  ifTrue:[Globals at: #AddSystemObjectsToGcCandidates put: false].
^false.
%

!-------------
! removal of _unsafesetoop method done later in filein

 
! end of bomlastconv.gs
commit
